#! /usr/bin/env python

## Copyright 1999-2001 by LivingLogic AG, Bayreuth, Germany.
## Copyright 1999-2001 by Walter Drwald
##
## All Rights Reserved
##
## Permission to use, copy, modify, and distribute this software and its documentation
## for any purpose and without fee is hereby granted, provided that the above copyright
## notice appears in all copies and that both that copyright notice and this permission
## notice appear in supporting documentation, and that the name of LivingLogic AG or
## the author not be used in advertising or publicity pertaining to distribution of the
## software without specific, written prior permission.
##
## LIVINGLOGIC AG AND THE AUTHOR DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
## INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
## LIVINGLOGIC AG OR THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
## DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
## IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
## IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""URL class.

This module contains only one useful variable: the URL class
"""

__version__ = tuple(map(int, "$Revision: 2.4 $"[11:-2].split(".")))
# $Source: /data/cvsroot/LivingLogic/xist/_xist/url.py,v $

import types, urlparse, urllib

from xist import helpers, options

urlparse.uses_relative.extend((None, "root"))

def _normalize(path):
	"""
	normalize the path by removing combinations of down/up
	and removing duplicate path markers.
	"""

	if path in ([], [u'.']): # shortcut for very common cases
		return path[:] # return a copy
	
	# the first element of absolute pathes '', preserve it
	wasAbsolutePath = len(path) and path[0] == u''
	path = [dir for dir in path if dir not in (u".", u"")]
	if wasAbsolutePath:
		path[:0] = [u'']

	# remove "foo/.." combinations
	# this works by using two pointers: 'i' for the up-dir part
	# and 'j' for the potential '..' part. After removing one
	# "foo/.." combination, 'i' and 'j' are set to the
	# predecessor or successor respectivly. Example:
	#	aa/bb/cc/../xx			i = 3; j = 4
	#	aa/bb/<None>/<None>/xx		i = 2; j = 5
	# This allows reduction of 'aa/bb/cc/../../' to 'aa/'
	i = 0; j = 1
	while j < len(path):
		if path[j] == u".." and path[i] != u"..": # up/down combination found
			path[i] = path[j] = None # None it out
			i -= 1; j += 1
			if i < 0:
				i = j; j += 1 # all leading dirs have been removed, so continue behind this part at 'j'
		else:
			i = j; j += 1 # nothing to do, move on one step
#### this is from urlparse.urljoin():  todo: implement here
## 	if len(segments) == 2 and segments[1] == '..' and segments[0] == '':
## 		segments[-1] = ''
## 	elif len(segments) >= 2 and segments[-1] == '..':
## 		segments[-2:] = ['']
	return [ dir for dir in path if dir is not None ] or [u"."]

class URL:
	"""
	This class represents XSC URLs.
	Every instance has the following instance variables:
	scheme -- The scheme (e.g. "http" or "ftp"); there's a special scheme "server" for server relative URLs
	server -- The server name
	port -- The port number
	path -- The path to the file as a list of strings
	file -- The filename including the extension
	parameters -- The parametes
	query -- The query
	fragment -- The fragment
	These variables form a URL in the following way
	<scheme>://<server>:<port>/<path>/<file>;<params>?<query>#<fragment>

	There is one additional feature supported by this class: The scheme
	root is interpreted as a file relative to the current directory and
	not the directory the parsed file is in.
	"""

	__safe = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_,.-:/"

	def __init__(self, url=None, scheme=None, server=None, port=None, path=None, file=None, parameters=None, query=None, fragment=None):
		# initialize the defaults
		self.scheme = self.server = self.port = None
		self.path = []
		self.file = self.parameters = self.query = self.fragment = None
		if url is None:
			pass
		elif type(url) in (types.StringType, types.UnicodeType):
			self.__fromString(url)
		elif isinstance(url, URL):
			self.scheme     = url.scheme
			self.server     = url.server
			self.port       = url.port
			self.path       = url.path[:] # make copy of list
			self.file       = url.file
			self.parameters = url.parameters
			self.query      = url.query
			self.fragment   = url.fragment
		else:
			raise ValueError("URL argument must be either a string or an URL or None")

		if scheme is not None:
			self.scheme = scheme
		if server is not None:
			self.server = server
		if port is not None:
			self.port = port
		if path is not None:
			self.path = map(helpers.unistr, path)
		if file is not None:
			self.file = file
		if parameters is not None:
			self.parameters = parameters
		if query is not None:
			self.query = query
		if fragment is not None:
			self.fragment = fragment

		self.__normalize()

	def __repr__(self):
		v = []
		if self.scheme is not None:
			v.append("scheme=" + repr(self.scheme))
		if self.server is not None:
			v.append("server=" + repr(self.server))
		if self.port is not None:
			v.append("port=" + repr(self.port))
		if self.path is not None:
			v.append("path=" + repr(self.path))
		if self.file is not None:
			v.append("file=" + repr(self.file))
		if self.parameters is not None:
			v.append("parameters=" + repr(self.parameters))
		if self.query is not None:
			v.append("query=" + repr(self.query))
		if self.fragment is not None:
			v.append("fragment=" + repr(self.fragment))
		return "URL(%s)" % ", ".join(v)

	def __add__(self, other):
		"""
		joins two URLs together. When the second URL is
		absolute (i.e. contains a scheme other than "server"
		or ""), you'll get a copy of the second URL.
		"""
		return self.__join(other)

	def __radd__(self, other):
		return URL(other).__join(self)

	__radd__.__doc__ = __add__.__doc__

	def clone(self):
		"""
		returns an identical clone of this URL.
		"""
		return URL(self)

	def isRemote(self):
		if not self.scheme: # may be None
			return 0
		elif self.scheme == u"server" and self.server == u"localhost":
			return 0
		else:
			return 1

	def __hasPath(self):
		"""returns true if path or file is given"""
		return self.path or self.scheme == "server" or self.file is not None

	def __requiresPath(self):
		a = self.server or self.file
		if self.path:
			return not a
		else:
			return not (a or self.fragment or self.parameters)

	
	def relativeTo(self, other):
		"""
		<par nointent>returns <self/> interpreted relative
		to <code>other</code>.</par>

		<par>Note that remote URLs won't be modified in any way,
		because although the file you've read might have been
		remote, the parsed XSC file that you output, probably
		isn't. The resulting URL won't have any path markers
		in it.</par>
		"""
		if not self.__hasPath():
			new = other + self # make eventually absolute URL
		else: # self has only some of params, query, fragment 
			new = self.clone()
		newpath = _normalize(new.path, removeAllMarkers=1)
		if not new.scheme:
			otherpath =_normalize(other.path, removeAllMarkers=1)
			if newpath == [u'.']: newpath = []
			if otherpath == [u'.']: otherpath = []
			while len(otherpath) and len(newpath) and otherpath[0] == newpath[0]:
				# throw away identical directories in
				# both paths (we don't have to go up
				# from file and down to path for these
				# identical directories)
				del otherpath[0]
				del newpath[0]
			# now for the rest of the path we have to go
			# up from file and down to path (the
			# directories for this are still in path)
			newpath[:0] = [u".."]*len(otherpath)
			if newpath == [] and new.__requiresPath():
				newpath = [u"."]
			if new.file == other.file:
				new.file = u""
		new.path = newpath
		return new

	def __cmp__(self, other):
		assert isinstance(other, URL), "URLs can only be compared with URLs"
		scheme1 = self.scheme
		if scheme1 is not None:
			scheme1 = scheme1.lower()
		scheme2 = other.scheme
		if scheme2 is not None:
			scheme2 = scheme2.lower()
		server1 = self.server
		if server1 is not None:
			server1 = server1.lower()
		server2 = other.server
		if server2 is not None:
			server2 = server2.lower()
		return cmp(scheme1, scheme2) or cmp(server1, server2) or cmp(self.port, other.port) or cmp(self.path, other.path) or cmp(self.file, other.file) or cmp(self.parameters, other.parameters) or cmp(self.query, other.query) or cmp(self.fragment, other.fragment)

	def open(self):
		return urllib.urlopen(self.__quote())

	def retrieve(self):
		return urllib.urlretrieve(self.__quote())

	def read(self):
		return self.open().read()

	def readlines(self):
		return self.open().readlines()

	def __fromString(self, url):
		(scheme, server, path, parameters, query, fragment) = urlparse.urlparse(url)
		scheme = helpers.unistr(scheme)
		server = helpers.unistr(server)
		path = helpers.unistr(path)
		parameters = helpers.unistr(parameters)
		query = helpers.unistr(query)
		fragment = helpers.unistr(fragment)
		if scheme == u"": # do we have a local file?
			if len(path):
				if path[0] == u"/": # this is a server relative URL
					path = path[1:] # drop the empty string in front of the first "/" ...
					scheme = u"server" # ... and use a special scheme for that
		elif scheme in (u"ftp", u"http", u"https"):
			if len(path) and len(server):
				path = path[1:] # the path from urlparse started with "/" too
		port = None
		pos = server.rfind(u":")
		if pos != -1:
			port = int(server[pos+1:])
			server = server[:pos]
		path = path.split(u"/")
		file = path[-1]
		if file in (u".", u".."):
			file = u""
		else:
			path = path[:-1]

		self.scheme = scheme or None
		self.server = server or None
		self.port = port
		self.path = path
		self.file = file
		self.parameters = parameters or None
		self.query = query or None
		self.fragment = fragment or None

	def asString(self):
		scheme = self.scheme or u""
		server = self.server or u""
		if self.port:
			server += u":" + str(self.port)
		path = self.path[:] # make a copy
		if scheme == u"server":
			scheme = u"" # remove our own private scheme name
			path[:0] = [u""] # make sure that there's a "/" at the start
		if not self.__requiresPath() and path == [u'.']:
			path = []
		file = self.file or u""
		path.append(file)
		url = urlparse.urlunparse((scheme, server, u"/".join(path), self.parameters or u"", self.query or u"", self.fragment or u""))
		return url

	def __quote(self):
		"""
		encodes the URL with % escapes
		"""
		url = self.asPlainString().encode("utf8")
		#v = []
		#for c in url:
		#	if c in self.__safe:
		#		v.append(c)
		#	else:
		#		v.append("%%%02x" % ord(c))
		#return "".join(v)
		return url

	def __join(base, url):
		"""Join a base URL and a possibly relative URL to form
		an absolute interpretation of the latter."""

		return URL(urlparse.urljoin(base.asString(), url.asString()))

	def __normalize(self):
		"""
		normalize the path by removing combinations of down/up
		and removing duplicate path markers.
		"""
		self.path = _normalize(self.path)

	def isRetrieve(self):
		"""
		returns if the file specified by the URL should be retrieved or not
		according to the "remoteness" of the URL and the retrieve options.
		"""
		remote = self.isRemote()
		if (options.retrieveremote and remote) or (options.retrievelocal and (not remote)):
			return 1
		else:
			return 0

def test_normalize():
	"""
	Tests whether '_normalize' returns the expected results.
	"""
	for input, output in (
		# Tuple format: input, output
		('', ''),
		('./', '.'),
		#( '/./', '/'), # _normalize does not handle absolute paths
		('xx', 'xx'),
		('xx/yy', 'xx/yy'),
		('xx/..', '.'),
		('xx/../.', '.'),
		('./xx/..', '.'),
		('./xx/../.', '.'),
		('xx/./..', '.'),
		('xx/yy/..', 'xx'),
		('xx//yy/../..', '.'),
		('xx/../..//', '..'),
		('xx/.././..', '..'),
		('xx/.', 'xx'),
		('./xx', 'xx'),
		#('/xx', '/xx'), # _normalize does not handle absolute paths
		#('/./xx', '/xx'), # dito

		('XX', 'XX'),
		('xx/XX', 'xx/XX'),
		('xx/XX/yy', 'xx/XX/yy'),
		('xx/XX/..', 'xx'),
		('xx/XX/../.', 'xx'),
		('xx/XX/./..', 'xx'),
		('./xx//XX/..', 'xx'),
		('./xx/XX/../.', 'xx'),
		('xx/XX/yy/..', 'xx/XX'),
		('xx/XX/yy/../..', 'xx'),
		('xx/yy/XX/../..', 'xx'),
		('xx/../XX/..', '.'),

		):
		#print `in_`, '\t', `out`
		test = "/".join(_normalize(input.split("/")))
		if test != output:
			raise str('Test failed. %r -> %r -> %r != %r' % (input, input.split('/'), test, output))
	print 'test passed: _normalize'

def test_url():
	"""
	Test whether instaziation, '+' and others have the expected effect.

	Check is done via comparing against another URL or a string; in the
	later case, the URL is converted using asString().
	"""
	import types
	for num, o_lhs, o_rhs in (
		(10, URL('.'), './'),
		(11, URL('./'), './'),
		(20, URL('..'), '../'),
		(21, URL('../'), '../'),
		(30, URL('http://aa/bb/cc.html'), 'http://aa/bb/cc.html'),
		(40, URL('http://aa/*BB/bb/cc.html'), 'http://aa/bb/cc.html'),
		(50, URL('/aa/bb/cc.html'), '/aa/bb/cc.html'),
		(60, URL('http:bb/cc/'), 'http:bb/cc/'),
		# these fail, should they?
		(69, URL('http:/bb/cc/'), 'http:/bb/cc/'),
		(70, URL('http:') + URL('/bb/cc/'), 'http:/bb/cc/'),
		# is this what is expected?
		(80, URL('http://test.com/index.html') + URL('impress.html'), 'http://test.com/impress.html'),
		(80, URL('/bb/cc/') + URL('http:'), 'http:'),
		(90, URL('#mark'), '#mark')
		):
		lhs = o_lhs
		rhs = o_rhs
		if type(lhs) != type(rhs):
			if type(rhs) is types.StringType:
				if isinstance(lhs, URL):
					lhs = o_lhs.asPlainString()
				else:
					lhs = str(lhs)
			elif type(lhs) is types.StringType:
				if isinstance(rhs, URL):
					rhs = o_rhs.asPlainString()
				else:
					rhs = str(rhs)
		if lhs != rhs:
			raise str('Test %s failed. %r (%r) != %r (%r)' % (num, o_lhs, lhs, o_rhs, rhs))
	print 'test passed: URL()'

def test_relativeTo():
	"""
	Test correctness of relativeTo().
	"""
	for From, to, should in (
		('./', './', './'),
		('cc.html', './', './'),
		('./cc.html', './', './'),
		('*/cc.html', '*/', './'),
		('*/cc.html', './', './'),

		('cc.html', '#mark', '#mark'),
		('*/cc.html', '*/#mark', './#mark'),
		('*/cc.html', '#mark', '#mark'),
		('*/cc.html', '*/cc.html#mark', '#mark'),
		('*/cc.html', '*/dd.html#mark', 'dd.html#mark'),
		('*/aa/bb/cc.html', '*/', '../../'),
		#('', '', ''),
		#('', '', ''),

		# relativeTo() does not really handle urls with scheme set
		('http://server/aa/bb.html', 'http://server/aa/cc.html', 'http://server/aa/cc.html'),

		# testing absolute pathes is worthless, since they have
		# a scheme = 'server' and thus are not really handled
		('/aa/bb.html', '/xx.html', '/xx.html'),
		):
		u = URL(to).relativeTo(URL(From)).asString()
		#print "\t".join((`URL(to)`, `URL(From)`, u, should))
		if u != should:
			raise str('Test failed:\n%r != %r\n(%r -> %r)' % (u, should, From, to))
	print 'test passed: relativeTo()'

test_input = urlparse.test_input

def test_url2():
	base = __empty = URL('')
	for line in test_input.splitlines():
		words = line.split()
		if not words:
			continue
		url = words[0]
		parts = URL(url)
		print '%-10s : %s' % (url, parts)
		abs = base + url
		if base is __empty:
			base = abs
		wrapped = '<URL:%s>' % abs.asString()
		print '%-10s = %s' % (url, wrapped)
		if len(words) == 3 and words[1] == '=':
			if wrapped != words[2]:
				print 'EXPECTED', words[2], '!!!!!!!!!!'

if __name__ == '__main__':
	test_normalize()
	test_url()
	test_relativeTo()
	test_url2()
